import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor

from torch_activation import register_activation

@register_activation
class MollifiedAbsoluteValue(nn.Module):
    r"""
    Applies the Mollified Absolute Value function:

    :math:`|x|_\epsilon = \sqrt{x^2 + \epsilon}`

    Args:
        epsilon (float, optional): Small constant for numerical stability. Default: 1e-6

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = MollifiedAbsoluteValue()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, epsilon=1e-6):
        super(MollifiedAbsoluteValue, self).__init__()
        self.epsilon = epsilon

    def forward(self, x) -> Tensor:
        return torch.sqrt(x.pow(2) + self.epsilon)


@register_activation
class SquarePlus(nn.Module):
    r"""
    Applies the SquarePlus function:

    :math:`\text{SquarePlus}(z) = \frac{1}{2} (z + |z|_\epsilon) = \frac{1}{2} (z + \sqrt{z^2 + \epsilon})`

    Args:
        epsilon (float, optional): Small constant for numerical stability. Default: 1e-6

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = SquarePlus()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, epsilon=1e-6):
        super(SquarePlus, self).__init__()
        self.epsilon = epsilon

    def forward(self, x) -> Tensor:
        return 0.5 * (x + torch.sqrt(x.pow(2) + self.epsilon))


@register_activation
class StepPlus(nn.Module):
    r"""
    Applies the StepPlus function:

    :math:`\text{StepPlus}(z) = \frac{1}{2} \left(1 + \frac{z}{|z|_\epsilon}\right)`

    Args:
        epsilon (float, optional): Small constant for numerical stability. Default: 1e-6

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = StepPlus()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, epsilon=1e-6):
        super(StepPlus, self).__init__()
        self.epsilon = epsilon

    def forward(self, x) -> Tensor:
        abs_x_eps = torch.sqrt(x.pow(2) + self.epsilon)
        return 0.5 * (1 + x / abs_x_eps)


@register_activation
class BipolarPlus(nn.Module):
    r"""
    Applies the BipolarPlus function:

    :math:`\text{BipolarPlus}(z) = \frac{z}{|z|_\epsilon}`

    Args:
        epsilon (float, optional): Small constant for numerical stability. Default: 1e-6

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = BipolarPlus()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, epsilon=1e-6):
        super(BipolarPlus, self).__init__()
        self.epsilon = epsilon

    def forward(self, x) -> Tensor:
        abs_x_eps = torch.sqrt(x.pow(2) + self.epsilon)
        return x / abs_x_eps


@register_activation
class LReLUPlus(nn.Module):
    r"""
    Applies the Leaky ReLU Plus function:

    :math:`\text{LReLUPlus}(z_i) = \frac{1}{2} (z_i + a_i z_i + |(1 - a_i) z_i|_\epsilon)`

    Args:
        negative_slope (float or Tensor, optional): Controls the angle of the negative slope. Default: 0.01
        epsilon (float, optional): Small constant for numerical stability. Default: 1e-6

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = LReLUPlus()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, negative_slope=0.01, epsilon=1e-6):
        super(LReLUPlus, self).__init__()
        self.negative_slope = negative_slope
        self.epsilon = epsilon

    def forward(self, x) -> Tensor:
        term = (1 - self.negative_slope) * x
        abs_term_eps = torch.sqrt(term.pow(2) + self.epsilon)
        return 0.5 * (x + self.negative_slope * x + abs_term_eps)


@register_activation
class vReLUPlus(nn.Module):
    r"""
    Applies the vReLU Plus function:

    :math:`\text{vReLUPlus}(z) = |z|_\epsilon`

    Args:
        epsilon (float, optional): Small constant for numerical stability. Default: 1e-6

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = vReLUPlus()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, epsilon=1e-6):
        super(vReLUPlus, self).__init__()
        self.epsilon = epsilon

    def forward(self, x) -> Tensor:
        return torch.sqrt(x.pow(2) + self.epsilon)


@register_activation
class SoftshrinkPlus(nn.Module):
    r"""
    Applies the Softshrink Plus function:

    :math:`\text{SoftshrinkPlus}(z) = z + \frac{1}{2} \left(\sqrt{(z - a)^2 + \epsilon} - \sqrt{(z + a)^2 + \epsilon}\right)`

    Args:
        lambda_val (float, optional): The lambda value for the Softshrink formulation. Default: 0.5
        epsilon (float, optional): Small constant for numerical stability. Default: 1e-6

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = SoftshrinkPlus()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, lambda_val=0.5, epsilon=1e-6):
        super(SoftshrinkPlus, self).__init__()
        self.lambda_val = lambda_val
        self.epsilon = epsilon

    def forward(self, x) -> Tensor:
        term1 = torch.sqrt((x - self.lambda_val).pow(2) + self.epsilon)
        term2 = torch.sqrt((x + self.lambda_val).pow(2) + self.epsilon)
        return x + 0.5 * (term1 - term2)


@register_activation
class PanPlus(nn.Module):
    r"""
    Applies the Pan Plus function:

    :math:`\text{PanPlus}(z) = -a + \frac{1}{2} \left(\sqrt{(z - a)^2 + \epsilon} + \sqrt{(z + a)^2 + \epsilon}\right)`

    Args:
        a (float, optional): The 'a' parameter in the Pan Plus formulation. Default: 0.5
        epsilon (float, optional): Small constant for numerical stability. Default: 1e-6

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = PanPlus()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a=0.5, epsilon=1e-6):
        super(PanPlus, self).__init__()
        self.a = a
        self.epsilon = epsilon

    def forward(self, x) -> Tensor:
        term1 = torch.sqrt((x - self.a).pow(2) + self.epsilon)
        term2 = torch.sqrt((x + self.a).pow(2) + self.epsilon)
        return -self.a + 0.5 * (term1 + term2)


@register_activation
class BReLUPlus(nn.Module):
    r"""
    Applies the Bounded ReLU Plus function:

    :math:`\text{BReLUPlus}(z) = \frac{1}{2} (1 + |z|_\epsilon - |z - 1|_\epsilon)`

    Args:
        epsilon (float, optional): Small constant for numerical stability. Default: 1e-6

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = BReLUPlus()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, epsilon=1e-6):
        super(BReLUPlus, self).__init__()
        self.epsilon = epsilon

    def forward(self, x) -> Tensor:
        abs_x_eps = torch.sqrt(x.pow(2) + self.epsilon)
        abs_x_minus_1_eps = torch.sqrt((x - 1).pow(2) + self.epsilon)
        return 0.5 * (1 + abs_x_eps - abs_x_minus_1_eps)


@register_activation
class SReLUPlus(nn.Module):
    r"""
    Applies the S-shaped ReLU Plus function:

    :math:`\text{SReLUPlus}(z_i) = a_i z_i + \frac{1}{2} (a_i - 1) (|z_i - t_i|_\epsilon - |z_i + t_i|_\epsilon)`

    Args:
        a (float or Tensor, optional): The 'a' parameter in the SReLU Plus formulation. Default: 0.5
        t (float or Tensor, optional): The 't' parameter in the SReLU Plus formulation. Default: 1.0
        epsilon (float, optional): Small constant for numerical stability. Default: 1e-6

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = SReLUPlus()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a=0.5, t=1.0, epsilon=1e-6):
        super(SReLUPlus, self).__init__()
        self.a = a
        self.t = t
        self.epsilon = epsilon

    def forward(self, x) -> Tensor:
        abs_x_minus_t_eps = torch.sqrt((x - self.t).pow(2) + self.epsilon)
        abs_x_plus_t_eps = torch.sqrt((x + self.t).pow(2) + self.epsilon)
        return self.a * x + 0.5 * (self.a - 1) * (abs_x_minus_t_eps - abs_x_plus_t_eps)


@register_activation
class HardTanhPlus(nn.Module):
    r"""
    Applies the HardTanh Plus function:

    :math:`\text{HardTanhPlus}(z) = \frac{1}{2} (|z + 1|_\epsilon - |z - 1|_\epsilon)`

    Args:
        epsilon (float, optional): Small constant for numerical stability. Default: 1e-6

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = HardTanhPlus()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, epsilon=1e-6):
        super(HardTanhPlus, self).__init__()
        self.epsilon = epsilon

    def forward(self, x) -> Tensor:
        abs_x_plus_1_eps = torch.sqrt((x + 1).pow(2) + self.epsilon)
        abs_x_minus_1_eps = torch.sqrt((x - 1).pow(2) + self.epsilon)
        return 0.5 * (abs_x_plus_1_eps - abs_x_minus_1_eps)


@register_activation
class HardshrinkPlus(nn.Module):
    r"""
    Applies the Hardshrink Plus function:

    :math:`\text{HardshrinkPlus}(z) = z \left(1 + \frac{1}{2} \left(\frac{z - a}{\sqrt{(z - a)^2 + \epsilon}} - \frac{z + a}{\sqrt{(z + a)^2 + \epsilon}}\right)\right)`

    Args:
        lambda_val (float, optional): The lambda value for the Hardshrink formulation. Default: 0.5
        epsilon (float, optional): Small constant for numerical stability. Default: 1e-6

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = HardshrinkPlus()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, lambda_val=0.5, epsilon=1e-6):
        super(HardshrinkPlus, self).__init__()
        self.lambda_val = lambda_val
        self.epsilon = epsilon

    def forward(self, x) -> Tensor:
        term1 = (x - self.lambda_val) / torch.sqrt((x - self.lambda_val).pow(2) + self.epsilon)
        term2 = (x + self.lambda_val) / torch.sqrt((x + self.lambda_val).pow(2) + self.epsilon)
        return x * (1 + 0.5 * (term1 - term2))


@register_activation
class MollifiedMeLUComponent(nn.Module):
    r"""
    Applies the Mollified MeLU Component function:

    :math:`\phi_{b_j c_j \text{Plus}}(z_i) = \frac{1}{2} \left(c_j - |z_i - b_j|_\epsilon + \sqrt{(c_j - |z_i - b_j|_\epsilon)^2 + \epsilon}\right)`

    Args:
        b (float, optional): The 'b' parameter in the MeLU formulation. Default: 0.0
        c (float, optional): The 'c' parameter in the MeLU formulation. Default: 1.0
        epsilon (float, optional): Small constant for numerical stability. Default: 1e-6

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = MollifiedMeLUComponent()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, b=0.0, c=1.0, epsilon=1e-6):
        super(MollifiedMeLUComponent, self).__init__()
        self.b = b
        self.c = c
        self.epsilon = epsilon

    def forward(self, x) -> Tensor:
        abs_x_minus_b_eps = torch.sqrt((x - self.b).pow(2) + self.epsilon)
        term = self.c - abs_x_minus_b_eps
        return 0.5 * (term + torch.sqrt(term.pow(2) + self.epsilon))


@register_activation
class TSAFPlus(nn.Module):
    r"""
    Applies the TSAF Plus function:

    :math:`\text{TSAFPlus}(z_i) = \frac{1}{4} \left(|z_i - a_i + c_i|_\epsilon + |z_i - a_i|_\epsilon + |z_i + b_i - c_i|_\epsilon - |z_i - b_i|_\epsilon\right)`

    Args:
        a (float, optional): The 'a' parameter in the TSAF formulation. Default: 0.5
        b (float, optional): The 'b' parameter in the TSAF formulation. Default: 0.5
        c (float, optional): The 'c' parameter in the TSAF formulation. Default: 1.0
        epsilon (float, optional): Small constant for numerical stability. Default: 1e-6

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = TSAFPlus()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a=0.5, b=0.5, c=1.0, epsilon=1e-6):
        super(TSAFPlus, self).__init__()
        self.a = a
        self.b = b
        self.c = c
        self.epsilon = epsilon

    def forward(self, x) -> Tensor:
        term1 = torch.sqrt((x - self.a + self.c).pow(2) + self.epsilon)
        term2 = torch.sqrt((x - self.a).pow(2) + self.epsilon)
        term3 = torch.sqrt((x + self.b - self.c).pow(2) + self.epsilon)
        term4 = torch.sqrt((x - self.b).pow(2) + self.epsilon)
        return 0.25 * (term1 + term2 + term3 - term4)


@register_activation
class ELUPlus(nn.Module):
    r"""
    Applies the ELU Plus function:

    :math:`\text{ELUPlus}(z) = \frac{1}{2} (z + |z|_\epsilon) + \frac{1}{2} \left(\frac{\exp(z) - 1}{a} + \sqrt{\left(\frac{\exp(z) - 1}{a}\right)^2 + \epsilon}\right)`

    Args:
        alpha (float, optional): The alpha value for the ELU formulation. Default: 1.0
        epsilon (float, optional): Small constant for numerical stability. Default: 1e-6

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = ELUPlus()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, alpha=1.0, epsilon=1e-6):
        super(ELUPlus, self).__init__()
        self.alpha = alpha
        self.epsilon = epsilon

    def forward(self, x) -> Tensor:
        term1 = 0.5 * (x + torch.sqrt(x.pow(2) + self.epsilon))
        elu_term = (torch.exp(x) - 1) / self.alpha
        term2 = 0.5 * (elu_term + torch.sqrt(elu_term.pow(2) + self.epsilon))
        return term1 + term2


@register_activation
class SwishPlus(nn.Module):
    r"""
    Applies the Swish Plus function:

    :math:`\text{SwishPlus}(z) = z \cdot \text{StepPlus}(z) = \frac{1}{2} \left(z + \frac{z^2}{|z|_\epsilon}\right)`

    Args:
        epsilon (float, optional): Small constant for numerical stability. Default: 1e-6

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = SwishPlus()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, epsilon=1e-6):
        super(SwishPlus, self).__init__()
        self.epsilon = epsilon

    def forward(self, x) -> Tensor:
        abs_x_eps = torch.sqrt(x.pow(2) + self.epsilon)
        return 0.5 * (x + (x.pow(2) / abs_x_eps))


@register_activation
class MishPlus(nn.Module):
    r"""
    Applies the Mish Plus function:

    :math:`\text{MishPlus}(z) = z \cdot \text{BipolarPlus}(\text{BipolarPlus}(z))`

    Args:
        epsilon (float, optional): Small constant for numerical stability. Default: 1e-6

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = MishPlus()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, epsilon=1e-6):
        super(MishPlus, self).__init__()
        self.epsilon = epsilon

    def forward(self, x) -> Tensor:
        # First BipolarPlus
        abs_x_eps = torch.sqrt(x.pow(2) + self.epsilon)
        bipolar1 = x / abs_x_eps
        
        # Second BipolarPlus
        abs_bipolar1_eps = torch.sqrt(bipolar1.pow(2) + self.epsilon)
        bipolar2 = bipolar1 / abs_bipolar1_eps
        
        return x * bipolar2


@register_activation
class LogishPlus(nn.Module):
    r"""
    Applies the Logish Plus function:

    :math:`\text{LogishPlus}(z) = z \cdot \ln(1 + \text{StepPlus}(z))`

    Args:
        epsilon (float, optional): Small constant for numerical stability. Default: 1e-6

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = LogishPlus()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, epsilon=1e-6):
        super(LogishPlus, self).__init__()
        self.epsilon = epsilon

    def forward(self, x) -> Tensor:
        abs_x_eps = torch.sqrt(x.pow(2) + self.epsilon)
        step_plus = 0.5 * (1 + x / abs_x_eps)
        return x * torch.log(1 + step_plus)


@register_activation
class SoftsignPlus(nn.Module):
    r"""
    Applies the Softsign Plus function:

    :math:`\text{SoftsignPlus}(z) = \frac{z}{1 + |z|_\epsilon}`

    Args:
        epsilon (float, optional): Small constant for numerical stability. Default: 1e-6

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = SoftsignPlus()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, epsilon=1e-6):
        super(SoftsignPlus, self).__init__()
        self.epsilon = epsilon

    def forward(self, x) -> Tensor:
        abs_x_eps = torch.sqrt(x.pow(2) + self.epsilon)
        return x / (1 + abs_x_eps)


@register_activation
class SignReLUPlus(nn.Module):
    r"""
    Applies the SignReLU Plus function:

    :math:`\text{SignReLUPlus}(z) = \frac{1}{2} (z + |z|_\epsilon) + \frac{z - |z|_\epsilon}{2 |1 - z|_\epsilon}`

    Args:
        epsilon (float, optional): Small constant for numerical stability. Default: 1e-6

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = SignReLUPlus()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, epsilon=1e-6):
        super(SignReLUPlus, self).__init__()
        self.epsilon = epsilon

    def forward(self, x) -> Tensor:
        abs_x_eps = torch.sqrt(x.pow(2) + self.epsilon)
        abs_1_minus_x_eps = torch.sqrt((1 - x).pow(2) + self.epsilon)
        term1 = 0.5 * (x + abs_x_eps)
        term2 = (x - abs_x_eps) / (2 * abs_1_minus_x_eps)
        return term1 + term2



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor

from torch_activation import register_activation

@register_activation
class SReLU(nn.Module):
    r"""
    Applies the S-shaped Rectified Linear Unit (SReLU) function:

    :math:`\text{SReLU}(z_i) = \begin{cases} 
    t^r_i + a^r_i(z_i - t^r_i), & z_i \geq t^r_i \\
    z_i, & t^r_i > z_i > t^l_i \\
    t^l_i + a^l_i(z_i - t^l_i), & z_i \leq t^l_i
    \end{cases}`

    where :math:`t^r_i`, :math:`t^l_i`, :math:`a^r_i`, and :math:`a^l_i` are trainable parameters.

    Args:
        init_tr (float, optional): Initial value for the right threshold parameter tr. Default: 1.0
        init_tl (float, optional): Initial value for the left threshold parameter tl. Default: 0.0
        init_ar (float, optional): Initial value for the right slope parameter ar. Default: 1.0
        init_al (float, optional): Initial value for the left slope parameter al. Default: 0.1
        fix_init_epochs (int, optional): Number of epochs to keep parameters fixed at initialization. Default: 0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = SReLU(init_tr=1.0, init_tl=0.0, init_ar=1.0, init_al=0.1)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, init_tr: float = 1.0, init_tl: float = 0.0, 
                 init_ar: float = 1.0, init_al: float = 0.1,
                 fix_init_epochs: int = 0):
        super(SReLU, self).__init__()
        self.tr = nn.Parameter(Tensor([init_tr]))
        self.tl = nn.Parameter(Tensor([init_tl]))
        self.ar = nn.Parameter(Tensor([init_ar]))
        self.al = nn.Parameter(Tensor([init_al]))
        
        # For tracking epochs if parameters should be fixed initially
        self.fix_init_epochs = fix_init_epochs
        self.current_epoch = 0
        self.init_tr = init_tr
        self.init_tl = init_tl
        self.init_ar = init_ar
        self.init_al = init_al

    def forward(self, x) -> Tensor:
        # Use fixed parameters during initial epochs if specified
        if self.current_epoch < self.fix_init_epochs and self.training:
            tr = self.init_tr
            tl = self.init_tl
            ar = self.init_ar
            al = self.init_al
        else:
            tr = self.tr
            tl = self.tl
            ar = self.ar
            al = self.al
        
        # Create masks for the three regions
        right_mask = x >= tr
        left_mask = x <= tl
        middle_mask = ~(right_mask | left_mask)
        
        # Initialize result tensor
        result = torch.zeros_like(x)
        
        # Apply the three piecewise linear functions
        if right_mask.any():
            result[right_mask] = tr + ar * (x[right_mask] - tr)
        
        if middle_mask.any():
            result[middle_mask] = x[middle_mask]
        
        if left_mask.any():
            result[left_mask] = tl + al * (x[left_mask] - tl)
        
        return result
    
    def train(self, mode=True):
        super(SReLU, self).train(mode)
        if mode:
            # Increment epoch counter when switching to training mode
            self.current_epoch += 1
        return self


@register_activation
class NActivation(nn.Module):
    r"""
    Applies the N-Activation function:

    :math:`\text{N-Activation}(z_i) = \begin{cases} 
    z_i - 2t_{i,min}, & z_i < t_{i,min} \\
    -z_i, & t_{i,min} \leq z_i \leq t_{i,max} \\
    z_i - 2t_{i,max}, & z_i > t_{i,max}
    \end{cases}`

    where :math:`t_{i,min} = \min(a_i, b_i)` and :math:`t_{i,max} = \max(a_i, b_i)`,
    and :math:`a_i` and :math:`b_i` are trainable parameters.

    Args:
        init_a (float, optional): Initial value for parameter a. Default: -0.5
        init_b (float, optional): Initial value for parameter b. Default: 0.5

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = NActivation(init_a=-0.5, init_b=0.5)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, init_a: float = -0.5, init_b: float = 0.5):
        super(NActivation, self).__init__()
        self.a = nn.Parameter(Tensor([init_a]))
        self.b = nn.Parameter(Tensor([init_b]))

    def forward(self, x) -> Tensor:
        # Calculate t_min and t_max
        t_min = torch.min(self.a, self.b)
        t_max = torch.max(self.a, self.b)
        
        # Create masks for the three regions
        left_mask = x < t_min
        middle_mask = (x >= t_min) & (x <= t_max)
        right_mask = x > t_max
        
        # Initialize result tensor
        result = torch.zeros_like(x)
        
        # Apply the three piecewise linear functions
        if left_mask.any():
            result[left_mask] = x[left_mask] - 2 * t_min
        
        if middle_mask.any():
            result[middle_mask] = -x[middle_mask]
        
        if right_mask.any():
            result[right_mask] = x[right_mask] - 2 * t_max
        
        return result


@register_activation
class ALiSA(nn.Module):
    r"""
    Applies the Adaptive Linearized Sigmoidal Activation (ALiSA) function:

    :math:`\text{ALiSA}(z_i) = \begin{cases} 
    a^r_i z_i - a^r_i + 1, & z_i \geq 1 \\
    z_i, & 1 > z_i > 0 \\
    a^l_i z_i, & z_i \leq 0
    \end{cases}`

    where :math:`a^r_i` and :math:`a^l_i` are trainable parameters.

    Args:
        init_ar (float, optional): Initial value for the right slope parameter ar. Default: 1.0
        init_al (float, optional): Initial value for the left slope parameter al. Default: 0.1

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = ALiSA(init_ar=1.0, init_al=0.1)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, init_ar: float = 1.0, init_al: float = 0.1):
        super(ALiSA, self).__init__()
        self.ar = nn.Parameter(Tensor([init_ar]))
        self.al = nn.Parameter(Tensor([init_al]))

    def forward(self, x) -> Tensor:
        # Create masks for the three regions
        right_mask = x >= 1
        left_mask = x <= 0
        middle_mask = ~(right_mask | left_mask)
        
        # Initialize result tensor
        result = torch.zeros_like(x)
        
        # Apply the three piecewise linear functions
        if right_mask.any():
            result[right_mask] = self.ar * x[right_mask] - self.ar + 1
        
        if middle_mask.any():
            result[middle_mask] = x[middle_mask]
        
        if left_mask.any():
            result[left_mask] = self.al * x[left_mask]
        
        return result


@register_activation
class LiSA(nn.Module):
    r"""
    Applies the Linearized Sigmoidal Activation (LiSA) function:

    :math:`\text{LiSA}(z_i) = \begin{cases} 
    a^r z_i - a^r + 1, & z_i \geq 1 \\
    z_i, & 1 > z_i > 0 \\
    a^l z_i, & z_i \leq 0
    \end{cases}`

    where :math:`a^r` and :math:`a^l` are fixed parameters.

    Args:
        ar (float, optional): Fixed value for the right slope parameter ar. Default: 1.0
        al (float, optional): Fixed value for the left slope parameter al. Default: 0.1

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = LiSA(ar=1.0, al=0.1)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, ar: float = 1.0, al: float = 0.1):
        super(LiSA, self).__init__()
        self.ar = ar
        self.al = al

    def forward(self, x) -> Tensor:
        # Create masks for the three regions
        right_mask = x >= 1
        left_mask = x <= 0
        middle_mask = ~(right_mask | left_mask)
        
        # Initialize result tensor
        result = torch.zeros_like(x)
        
        # Apply the three piecewise linear functions
        if right_mask.any():
            result[right_mask] = self.ar * x[right_mask] - self.ar + 1
        
        if middle_mask.any():
            result[middle_mask] = x[middle_mask]
        
        if left_mask.any():
            result[left_mask] = self.al * x[left_mask]
        
        return result